from ...crazy import CrazyDragon

from .constants import Kp, Kd, g

from numpy import zeros

from time import sleep



def landing_supporter( cf: CrazyDragon, option=1, dt=0.1, step=0.03 ):

    des     = zeros(3)
    acc_cmd = zeros(3)
    P_pos   = zeros(3)
    D_pos   = zeros(3)

    print( 'landing supporter' )

    pos = cf.pos
    vel = cf.vel

    g = 9.81

    des[:] = pos

    for _ in range( 10 ):

        P_pos[:] = des - pos
        D_pos[:] = vel

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] -= D_pos * Kd

        acc_cmd[2] = g

        cf.command[:] = acc_cmd

        sleep( dt )

    if option:
        des[:] = pos
    else:
        des[:2] = 0
        des[2:] = pos[2]

    for _ in range( 30 ):

        P_pos[:] = des - pos
        D_pos[:] = vel

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] -= D_pos * Kd

        g -= step

        acc_cmd[2] = g

        cf.command[:] = acc_cmd

        sleep( dt )

    print( 'land' )


import matplotlib.pyplot as plt



def plot_acc_pos_cmd(acc, acccmd, vel, pos, posref, _len):

    plt.figure(figsize=(14,14))

    plt.subplot(331)
    plt.plot(acc[0,:_len], label='realtime acceleration x')
    plt.plot(acccmd[0,:_len], label='command input x')
    plt.legend()
    plt.grid()

    plt.subplot(332)
    plt.plot(acc[1,:_len], label='realtime acceleration y')
    plt.plot(acccmd[1,:_len], label='command input y')
    plt.legend()
    plt.grid()

    plt.subplot(333)
    plt.plot(acc[2,:_len], label='realtime acceleration z')
    plt.plot(acccmd[2,:_len], label='command input z')
    plt.legend()
    plt.grid()

    ## velocity is reversed
    plt.subplot(334)
    plt.plot(vel[0,:_len], label='realtime velocity x')
    plt.legend()
    plt.grid()

    plt.subplot(335)
    plt.plot(vel[1,:_len], label='realtime velocity y')
    plt.legend()
    plt.grid()

    plt.subplot(336)
    plt.plot(vel[2,:_len], label='realtime velocity z')
    plt.legend()
    plt.grid()

    plt.subplot(337)
    plt.plot(pos[0,:_len], label='realtime position x')
    plt.plot(posref[0,:_len], label='destination position x')
    plt.legend()
    plt.grid()

    plt.subplot(338)
    plt.plot(pos[1,:_len], label='realtime position y')
    plt.plot(posref[1,:_len], label='destination position y')
    plt.legend()
    plt.grid()

    plt.subplot(339)
    plt.plot(pos[2,:_len], label='realtime position z')
    plt.plot(posref[2,:_len], label='destination position z')
    plt.legend()
    plt.grid()

    plt.show()


def plot_thrust(thrust, thrustcmd, _len):

    plt.figure(figsize=(8,8))

    plt.plot(thrust[:_len], label='realtime thrust')
    plt.plot(thrustcmd[:_len], label='reference thrust')

    plt.legend()
    plt.grid()

    plt.show()


def plot_att(att, attcmd, _len):

    plt.figure(figsize=(14,14))
    
    plt.subplot(311)
    plt.plot(att[0,:_len], label='realtime euler x')
    plt.plot(attcmd[0,:_len], label='realtime euler input x')
    plt.legend()
    plt.grid()

    plt.subplot(312)
    plt.plot(att[1,:_len], label='realtime euler y')
    plt.plot(attcmd[1,:_len], label='realtime euler input y')
    plt.legend()
    plt.grid()

    plt.subplot(313)
    plt.plot(att[2,:_len], label='realtime euler z')
    plt.plot(attcmd[2,:_len], label='realtime euler input z')
    plt.legend()
    plt.grid()

    plt.show()


from crazy import CrazyDragon

import time

from cflib.crazyflie.log        import LogConfig
from cflib.crazyflie.syncLogger import SyncLogger



orientation_std_dev = 8.0e-3


def adjust_orientation_sensitivity( cf: CrazyDragon ):
    cf.param.set_value('locSrv.extQuatStdDev', orientation_std_dev)


def activate_kalman_estimator( cf: CrazyDragon ):
    cf.param.set_value('stabilizer.estimator', '2')

    # Set the std deviation for the quaternion data pushed into the
    # kalman filter. The default value seems to be a bit too low.
    cf.param.set_value('locSrv.extQuatStdDev', 0.06)


def reset_estimator( cf: CrazyDragon ):
    cf.param.set_value('kalman.resetEstimation', '1')
    time.sleep(0.1)
    cf.param.set_value('kalman.resetEstimation', '0')

    # time.sleep(1)
    wait_for_position_estimator( cf )


def wait_for_position_estimator( cf: CrazyDragon ):
    print('Waiting for estimator to find position...')

    log_config = LogConfig(name='Kalman Variance', period_in_ms=10)
    log_config.add_variable('kalman.varPX', 'FP16')
    log_config.add_variable('kalman.varPY', 'FP16')
    log_config.add_variable('kalman.varPZ', 'FP16')

    

    var_y_history = [1000] * 10
    var_x_history = [1000] * 10
    var_z_history = [1000] * 10

    threshold = 0.001

    with SyncLogger( cf, log_config ) as logger:
        for log_entry in logger:
            data = log_entry[1]

            var_x_history.append( data['kalman.varPX'] )
            var_x_history.pop( 0 )
            var_y_history.append( data['kalman.varPY'] )
            var_y_history.pop( 0 )
            var_z_history.append( data['kalman.varPZ'] )
            var_z_history.pop( 0 )

            min_x = min( var_x_history )
            max_x = max( var_x_history )
            min_y = min( var_y_history )
            max_y = max( var_y_history )
            min_z = min( var_z_history )
            max_z = max( var_z_history )

            if ( max_x - min_x ) < threshold and (
                    max_y - min_y ) < threshold and (
                    max_z - min_z ) < threshold:
                break


from crazy import CrazyDragon

from threading import Thread

from numpy import zeros, array

from time  import sleep

from control import alpha

from .visualizer import *



class Recorder( Thread ):


    def __init__(self, cf: CrazyDragon, commander, n=10000):
        ## for threading
        super().__init__()

        self.record_length = 0
        self.recording     = True

        self.guidance_start_idx = -1
        self.guidance_end_idx   = -1

        self.cf = cf

        ## callback functions
        self.record_callback = {
            'acccmd': array_type_data_callback,
            'vel'   : array_type_data_callback,
            'pos'   : array_type_data_callback,
            'posref': array_type_data_callback,
            'att'   : array_type_data_callback,
            'cmd'   : array_type_data_callback,
            'thrust': float_type_data_callback
        }
        ## data storage
        self.record_datastrg = {
            'acc'   : zeros((3,n)),
            'acccmd': zeros((3,n)),
            'vel'   : zeros((3,n)),
            'pos'   : zeros((3,n)),
            'posref': zeros((3,n)),
            'att'   : zeros((3,n)),
            'cmd'   : zeros((4,n)),
            'thrust': zeros((1,n))
        }
        ## realtime data
        self.realtime_data = {
            'acc'   : cf.acc,
            'acccmd': cf.command,
            'vel'   : cf.vel,
            'pos'   : cf.pos,
            'posref': cf.destination,
            'att'   : cf.att,
            'cmd'   : commander.command,
            'thrust': commander.thrust
        }

    
    def run(self):

        sleep(0.1)

        cf = self.cf

        while self.recording:

            self.record_datastrg['acc'][:,self.record_length] = self.realtime_data['acc']

            for key, callback in self.record_callback.items():

                datastrg = self.record_datastrg[key]
                data     = self.realtime_data[key]

                callback( datastrg, data, self.record_length )

            self.record_length += 1

            sleep(0.05)

    
    def stop_record(self):

        self.recording = False

    
    def guidance_start(self):

        self.guidance_start_idx = self.record_length


    def guidance_end(self):

        self.guidance_end_idx = self.record_length

    
    def join(self):

        super().join()

        _len = self.record_length

        acc    = self.record_datastrg['acc']
        acccmd = self.record_datastrg['acccmd']

        vel    = self.record_datastrg['vel']

        pos    = self.record_datastrg['pos']
        posref = self.record_datastrg['posref']

        att    = self.record_datastrg['att']
        cmd    = self.record_datastrg['cmd']
        thrust = self.record_datastrg['thrust']

        plot_acc_pos_cmd(acc, acccmd, vel, pos, posref, _len)
        plot_thrust(thrust[0,:], cmd[3,:]*alpha, _len)
        plot_att(att, cmd[:3,:], _len)


def array_type_data_callback(datastrg, data, i):

    datastrg[:,i] = array(data)


def float_type_data_callback(datastrg, data, i):

    datastrg[0,i] = data